unit MeshHierarchyUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, ComCtrls, Menus, AppEvnts, StdCtrls, Grids, ValEdit, 
  D3DX9, DXFile, D3Dpart;

type
  CMyD3DApplication2 = class(CMyD3DApplication) end;

  THierarchyForm = class(TForm)
    DisplayPanel: TPanel;
    Splitter1: TSplitter;
    MainMenu1: TMainMenu;
    StatusBar: TStatusBar;
    File1: TMenuItem;
    Exit1: TMenuItem;
    Open1: TMenuItem;
    Saveas1: TMenuItem;
    N1: TMenuItem;
    About1: TMenuItem;
    OpenDialog: TOpenDialog;
    ApplicationEvents1: TApplicationEvents;
    Panel1: TPanel;
    Splitter2: TSplitter;
    reeView1: TMenuItem;
    ExpandAll1: TMenuItem;
    Collapseall1: TMenuItem;
    Open2: TMenuItem;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TreeView: TTreeView;
    TabSheet2: TTabSheet;
    GlobalPreValueListEditor: TValueListEditor;
    GlobalPostValueListEditor: TValueListEditor;
    Bevel1: TBevel;
    MeshValueListEditor: TValueListEditor;
    InfoPanel: TPanel;
    InfoText: TStaticText;
    Edit1: TEdit;
    StaticText1: TStaticText;
    Bevel2: TBevel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Options1: TMenuItem;
    AutoGenerateChildUserData: TMenuItem;
    SaveDialog: TSaveDialog;
    procedure Open2Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure OnToggleFullscreen(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ExpandAll1Click(Sender: TObject);
    procedure Collapseall1Click(Sender: TObject);
    procedure TreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure Saveas1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure Edit1Change(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure AutoGenerateChildUserDataClick(Sender: TObject);
    procedure DisplayPanelResize(Sender: TObject);
  private
    { Private declarations }
    fMyApp: CMyD3DApplication2;
    fFileName: String;
    fAnimationController: ID3DXAnimationController;
    fSelectedFrame: Pointer;
    procedure LoadMesh(aName: String; UseLoadExtentions: Boolean = False);
    procedure DisposeMesh;
    procedure FillTreeView;
  public
    { Public declarations }
  end;

var
  HierarchyForm: THierarchyForm;

///////////////////////////////////////////////////////////////////////////

const
  sAppName = 'Clootie - Mesh Hierarchy';

implementation

uses HRoutines;

{$R *.dfm}

procedure THierarchyForm.LoadMesh(aName: String; UseLoadExtentions: Boolean = False);
var
  ah: TAllocateHierarchy;
  ld: TD3DXLoadUserData;
  frame: PD3DXFrame;
  hr: HRESULT;
begin
  DisposeMesh; // Clean up

  fFileName:= aName;
  fAnimationController:= nil;

  ah:= TAllocateHierarchy.Create;
  ld:= TD3DXLoadUserData.Create;

  ////////////////////////////////////////////////////////////
  if UseLoadExtentions then
    hr:= D3DXLoadMeshHierarchyFromX(PChar(aName), D3DXMESH_MANAGED, fMyApp.m_pd3dDevice,
      ah, ld,  frame, fAnimationController)
  else
    hr:= D3DXLoadMeshHierarchyFromX(PChar(aName), D3DXMESH_MANAGED, fMyApp.m_pd3dDevice,
      ah, nil, frame, fAnimationController);
  ////////////////////////////////////////////////////////////

  if Failed(hr) then
    Application.MessageBox('Error during loading of .x file', PChar(sAppName));

  ld.Free;
  ah.Free;

  FillTreeView;
end;

procedure THierarchyForm.Saveas1Click(Sender: TObject);
var
  sd: TD3DXSaveUserData;
begin
  if SaveDialog.Execute then
  begin
    sd:= TD3DXSaveUserData.Create;
    D3DXSaveMeshHierarchyToFile(PChar(SaveDialog.FileName), DXFILEFORMAT_TEXT,
      RootFrame, fAnimationController, sd);
    sd.Free;
  end;
end;

procedure THierarchyForm.DisposeMesh;
var
  ah: TAllocateHierarchy;
begin
  fAnimationController:= nil;

  if (RootFrame <> nil) then
  begin
    ah:= TAllocateHierarchy.Create;
    D3DXFrameDestroy(RootFrame, ah);
    ah.Free;
    RootFrame := nil;
  end;
  GlobalPreValueListEditor.Strings.Clear;
  GlobalPostValueListEditor.Strings.Clear;
  MeshValueListEditor.Strings.Clear;
end;

procedure THierarchyForm.FillTreeView;
  procedure AddItem(Frame: PD3DXFrame; node: TTreeNode);
  var
    ss: String;
  begin
    //todo: Add some other information from Frame to node...
    if (Frame.pMeshContainer <> nil) then
    begin
      ss:= Frame.pMeshContainer.Name;
      node.Text:= Format('%s (%s)', [node.Text, ss]);
    end;

    // Parse next levels
    Frame:= Frame.pFrameFirstChild;
    while (Frame <> nil) do
    begin
      AddItem(Frame,
        TreeView.Items.AddChildObject(node, Frame.Name, Frame));
      Frame := Frame.pFrameSibling;
    end;
  end;
var
  CurrentFrame: PD3DXFrame;
begin
  TreeView.Items.Clear;

  CurrentFrame:= RootFrame;
  while (CurrentFrame <> nil) do
  begin
    AddItem(CurrentFrame,
      TreeView.Items.AddChildObject(nil, CurrentFrame.Name, CurrentFrame));
    CurrentFrame := CurrentFrame.pFrameSibling;
  end;
end;

procedure THierarchyForm.TreeViewChange(Sender: TObject; Node: TTreeNode);
var
  s1, s2: String;
  i: Integer;
  Frame: PD3DXFrameDerived;
begin
  i:= -1;
  if Node <> nil then
  begin
    Frame:= Node.Data;
    Assert(Frame <> nil);
    s1:= Frame.Name;
    if Frame.pMeshContainer <> nil then
    begin
      s2:= Frame.pMeshContainer.Name;
      i:= Ord(Frame.pMeshContainer.MeshData._Type);
    end;
    InfoText.Caption:= Format('%s (%s)'#10#13'%d', [s1, s2, i]);

    fSelectedFrame:= Frame;
    Edit1.Text:= IntToStr(Frame.UserData);
  end;
end;

procedure THierarchyForm.Edit1Change(Sender: TObject);
var
  Frame: PD3DXFrameDerived;
begin
  if not Assigned(fSelectedFrame) then Exit;
  Frame:= fSelectedFrame;
  Frame.UserData:= StrToIntDef(Edit1.Text, 0);
end;

procedure THierarchyForm.Open2Click(Sender: TObject);
begin
  if OpenDialog.Execute then LoadMesh(OpenDialog.FileName);
end;

procedure THierarchyForm.Open1Click(Sender: TObject);
begin
  if OpenDialog.Execute then LoadMesh(OpenDialog.FileName, True);
end;

//-----------------------------------------------------------------------------
// Name: FullScreenWndProc()
// Desc: The WndProc funtion used when the app is in fullscreen mode. This is
//       needed simply to trap the ESC key.
//-----------------------------------------------------------------------------
function FullScreenWndProc(hWnd: HWND; msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT; stdcall;
begin
  if (msg = WM_CLOSE) then
  begin
    // User wants to exit, so go back to windowed mode and exit for real
    HierarchyForm.OnToggleFullScreen(nil);
    PostMessage(Application.Handle, WM_CLOSE, 0, 0);
  end;

  if (msg = WM_SETCURSOR) then SetCursor(0);

  // User wants to leave fullscreen mode
  if (msg = WM_KEYUP) and (wParam = VK_ESCAPE) then
    HierarchyForm.OnToggleFullScreen(nil);

  Result:= DefWindowProc(hWnd, msg, wParam, lParam);
end;

procedure THierarchyForm.FormShow(Sender: TObject);
{$WRITEABLECONST ON}
const
  wndClass: TWndClass = (
    style: CS_HREDRAW or CS_VREDRAW;
    lpfnWndProc: @FullScreenWndProc; cbClsExtra: 0; cbWndExtra: 0; hInstance: 0;
    hIcon: 0; hCursor: 0; hbrBackground: 0; lpszMenuName: nil;
    lpszClassName: 'Fullscreen Window'
  );
{$WRITEABLECONST OFF}
begin
  OpenDialog.InitialDir:= ExtractFileDir(ParamStr(0));
  SaveDialog.InitialDir:= ExtractFileDir(ParamStr(0));

  fMyApp:= CMyD3DApplication2.Create;
  fMyApp.fAppForm:= Self;
  fMyApp.m_hwndRenderWindow:= DisplayPanel.Handle; // Save static reference to the render window

  // Register a class for a fullscreen window
  (*
  WNDCLASS wndClass = { CS_HREDRAW | CS_VREDRAW, FullScreenWndProc, 0, 0, NULL,
                        NULL, NULL, (HBRUSH)GetStockObject(WHITE_BRUSH), NULL,
                        _T("Fullscreen Window") };
  *)
  wndClass.hbrBackground:= GetStockObject(WHITE_BRUSH);
  Windows.RegisterClass(wndClass);

  // We create the fullscreen window (not visible) at startup, so it can
  // be the focus window.  The focus window can only be set at CreateDevice
  // time, not in a Reset, so ToggleFullscreen wouldn't work unless we have
  // already set up the fullscreen focus window.
  fMyApp.m_hwndRenderFullScreen:= CreateWindow('Fullscreen Window', nil,
                                      WS_POPUP, Integer(CW_USEDEFAULT),
                                      Integer(CW_USEDEFAULT), 100, 100,
                                      Application.Handle, 0, 0, nil);

  fMyApp.AssignWindowsHandles;
  fMyApp.AppCreate(HInstance);

  LoadMesh('tiny.x', True); //todo: Modify this later

  ExpandAll1Click(nil);
end;

//-----------------------------------------------------------------------------
// Name: OnToggleFullScreen()
// Desc: Called when user toggles the fullscreen mode
//-----------------------------------------------------------------------------
procedure THierarchyForm.OnToggleFullscreen(Sender: TObject);
begin
  fMyApp.ToggleFullscreen;
end;

procedure THierarchyForm.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure THierarchyForm.ApplicationEvents1Idle(Sender: TObject;
  var Done: Boolean);
var
  strStatsPrev: String;
begin
  Done:= True;

  strStatsPrev:= fMyApp.m_strFrameStats;

  // Do not render if the app is minimized
  if IsIconic(Application.Handle) then Exit;

  // Update and render a frame
  if fMyApp.IsReady then
  begin
    fMyApp.Render3DEnvironment;
    if (strStatsPrev <> fMyApp.m_strFrameStats) then
      StatusBar.Panels[0].Text:= fMyApp.m_strFrameStats;
  end;

  // Keep requesting more idle time
  Done:= False;
end;

procedure THierarchyForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  DisposeMesh;

  fMyApp.Cleanup3DEnvironment;
  fMyApp.FinalCleanup;
  FreeAndNil(fMyApp);
end;

procedure THierarchyForm.ExpandAll1Click(Sender: TObject);
begin
  if Assigned(TreeView) and (TreeView.Items.Count > 0) then
    TreeView.Items[0].Expand(True);
end;

procedure THierarchyForm.Collapseall1Click(Sender: TObject);
begin
  if Assigned(TreeView) and (TreeView.Items.Count > 0) then
    TreeView.Items[0].Collapse(True);
end;

procedure THierarchyForm.About1Click(Sender: TObject);
begin
  Application.MessageBox('This demo shows how to load / parse / save mesh hierarchy.'#10 +
    'Also it shows how to load / save USER DATA embedded in .x file using D3DX9 functionality.'#10#10 +
    'This and other demos can be found at http://clootie.narod.ru',
    'Hierarchy Loading & Saving Demo', MB_ICONINFORMATION or MB_OK);
end;

procedure THierarchyForm.AutoGenerateChildUserDataClick(Sender: TObject);
begin
  with Sender as TMenuItem do Checked:= not Checked;
  AutoGenerateUserData:= (Sender as TMenuItem).Checked;
end;

procedure THierarchyForm.DisplayPanelResize(Sender: TObject);
begin
  if Assigned(fMyApp) and fMyApp.IsReady then
  begin
    fMyApp.Pause(True);
    fMyApp.MsgProc(0, WM_EXITSIZEMOVE, 0, 0);
  end;
end;

end.

